#pragma  once

#include "CImg.h"
#include "ImageEx.h"
#include "Timer.h"
#include "TMatrix.h"

using namespace cimg_library;

template <typename TypeOfKernelAndOutput>
class CKernel1D
{
	public:
		TypeOfKernelAndOutput*	data;
		int						center;
		int						size;

	public:
		CKernel1D(const CKernel1D& b)
		{
			center = b.center;
			size = b.size;
			if (b.data)
			{
				data = new TypeOfKernelAndOutput[size];
				memcpy(data,b.data,sizeof(TypeOfKernelAndOutput)*size);
			}
			else
			{
				data = NULL;
			}
		}

		CKernel1D()
		{
			center = size = -1;
			data = NULL;
		}

		virtual size_t GetMemoryUsage() const
		{
			return sizeof(int) * 2 + sizeof(TypeOfKernelAndOutput*) + sizeof(TypeOfKernelAndOutput) * size;
		}

		void InitBySize(int Size, int Center)
		{
			center	= Center;
			size	= Size;
			if (data)
			{
				delete[] data;
			}
			data = new TypeOfKernelAndOutput[Size];
		}

		void InitByCenter(int halfSize)
		{
			center = halfSize;
			if (data && (size == center+center+1))
			{
				return;
			}
			size = center+center+1;
			if (data)
			{
				delete[] data;
			}
			data = new TypeOfKernelAndOutput[size];
		}

		void InitBySigma(TypeOfKernelAndOutput sigma)
		{
			if (data)
			{
				delete[] data;
			}
			data = NULL;
			InitByCenter((int)ceil(2.0 * sigma + 1e-6));
		}

		~CKernel1D()
		{
			if (data)
			{
				delete[] data;
			}
		}

	public:
		// (dstData) += (srcData) *convolve* (this kernel)
		template <typename TypeOfInput>
		void _convolve(unsigned int dim, TypeOfKernelAndOutput* dstData, const TypeOfInput* srcData, unsigned int dstJumpOffset, unsigned int srcJumpOffset, unsigned int mskJumpOffset) const
		{
			TypeOfKernelAndOutput t;
			TypeOfKernelAndOutput* const pDstHead = dstData;

			const TypeOfKernelAndOutput* const pDstTail = dstData + dim * dstJumpOffset;

			const TypeOfKernelAndOutput* const pMskTail = data + size;
			const TypeOfKernelAndOutput* pMsk;

			const TypeOfInput* const pSrcHead = srcData;
			const TypeOfInput* const pSrcTail = srcData + dim * srcJumpOffset;
			const TypeOfInput* pSrc;

			mskJumpOffset *= srcJumpOffset;
			for (dstData=pDstHead; dstData<pDstTail; dstData+=dstJumpOffset)
			{
				pSrc	= srcData - center * mskJumpOffset;
				t		= (TypeOfKernelAndOutput)0;
				for (pMsk=data; pMsk<pMskTail; ++pMsk)
				{
					if (pSrc<pSrcHead)
						t += ((TypeOfKernelAndOutput)(*pSrcHead)) * (*pMsk);
					else if (pSrc>=pSrcTail)
						t += ((TypeOfKernelAndOutput)(*(pSrcTail-1))) * (*pMsk);
					else
						t += ((TypeOfKernelAndOutput)(*pSrc)) * (*pMsk);
					pSrc += mskJumpOffset;
				}
				*dstData = t;
				srcData += srcJumpOffset;
			}
		}

	public:
		static void GenerateSingle(CKernel1D<TypeOfKernelAndOutput>& DstKernel1D, TypeOfKernelAndOutput scale = (TypeOfKernelAndOutput)1)
		{
			DstKernel1D.InitByCenter(0);
			DstKernel1D.data[0] = scale;
		}
		static void GenerateGaussian(CKernel1D<TypeOfKernelAndOutput>& DstKernel1D, TypeOfKernelAndOutput sigma)
		{
			DstKernel1D.InitBySigma(sigma);
			for ( int i = 0; i <= DstKernel1D.center; ++i )
			{
				DstKernel1D.data[DstKernel1D.center+i] = DstKernel1D.data[DstKernel1D.center-i] =
				//CMyMath::gaussian<TypeOfKernelAndOutput>((TypeOfKernelAndOutput)i,0.0f,sigma);
				exp((i*i)/(-2.0f*sigma*sigma)) / (sqrt(M_PI+M_PI)*sigma);
			}
		}
};


class CExternalTools
{
	public:
		// only for color image
		static void ResizeImage(CImg<double> &Dst, CImg<double> &Src, double Scale)
		{
			CImg<double>	Tmp;
			int				SrcW	= Src.dimx( );
			int				SrcH	= Src.dimy( );
			int				DstW	= int(SrcW * Scale + 0.5);
			int				DstH	= int(SrcH * Scale + 0.5);

			Dst.assign(DstW, DstH, Src.dimz( ), Src.dimv( ), 0);
			Tmp.assign(SrcH, DstW, Src.dimz( ), Src.dimv( ), 0);
			ImageResize1DTrans(Src, Tmp);
			ImageResize1DTrans(Tmp, Dst);
		}

		static void ImageResize1DTrans(CImg<double> &Src, CImg<double> &Dst)
		{
			vector<int>			SrcIndex;
			vector<int>			DstIndex;
			vector<double>		Alpha;

			int		SrcW		= Src.dimx( );
			int		SrcH		= Src.dimy( );
			int		DstW		= Dst.dimx( );
			int		DstH		= Dst.dimy( );
			double	Scale		= 1.0f * DstH / SrcW;
			double	InvScale	= 1.0f * SrcW / DstH;

			Alpha.clear( );
			SrcIndex.clear( );
			DstIndex.clear( );
			for ( int t = 0; t < DstH; t++ )
			{
				double	SrcPos_Low	= t * InvScale;
				double	SrcPos_High	= SrcPos_Low + InvScale;
				int		LowPos		= int(SrcPos_Low) + 1;
				int		HighPos		= int(SrcPos_High);

				if (LowPos - SrcPos_Low > 1e-3)
				{
					SrcIndex.push_back(LowPos - 1);
					DstIndex.push_back(t*DstW);
					Alpha.push_back((LowPos - SrcPos_Low) * Scale);
				}
				for ( int p = LowPos; p < HighPos; p++ )
				{
					SrcIndex.push_back(p);
					DstIndex.push_back(t*DstW);
					Alpha.push_back(Scale);
				}				
				if (SrcPos_High - HighPos > 1e-3)
				{
					SrcIndex.push_back(HighPos);
					DstIndex.push_back(t*DstW);
					Alpha.push_back((SrcPos_High - HighPos) * Scale);
				}
			}

			for ( int v = 0; v < Src.dimv( ); v++ )
			{
				for ( int z = 0; z < Src.dimz( ); z++ )
				{
					double	*pSrc	= Src.data + v * Src.dimx( ) * Src.dimy( ) * Src.dimz( ) + z * Src.dimx( ) * Src.dimy( );
					double	*pDst	= Dst.data + v * Dst.dimx( ) * Dst.dimy( ) * Dst.dimz( ) + z * Dst.dimx( ) * Dst.dimy( );
					for ( int t = 0; t < SrcH; t++ )
					{
						for ( int p = 0; p < (int) Alpha.size( ); p++ )
						{
							pDst[DstIndex[p]]	+= Alpha[p] * pSrc[SrcIndex[p]];
						}
						pSrc	+= Src.dimx( );
						pDst	+= 1;
					}
				}
			}
		}
		
		template <typename T>
		static void myLoadImage(string strImageFileName, CImg<T>& dstImage)
		{
			CImageEx *ptrTmpImg = new CImageEx();
			ptrTmpImg->Load(strImageFileName.c_str());
			ptrTmpImg->ConvertToARGB32Bits();
			RGBQUAD *pBits = (RGBQUAD*)(ptrTmpImg->GetBits());
			int w = ptrTmpImg->GetWidth();
			int h = ptrTmpImg->GetHeight();
			dstImage.assign(w,h,1,3);
			int x,y;
			for ( y = 0; y < h; ++y )
			{
				for ( x = 0; x < w; ++x )
				{
					dstImage.at(x,y,0,0) = (T)pBits->rgbRed;
					dstImage.at(x,y,0,1) = (T)pBits->rgbGreen;
					dstImage.at(x,y,0,2) = (T)pBits->rgbBlue;
					++pBits;
				}
			}
			ptrTmpImg->FreeBits();
			delete ptrTmpImg;
		}

		template <typename T>
		static void mySaveImage(string strImageFileName, CImg<T>& srcImage)
		{
			CImageEx *ptrTmpImg = new CImageEx();
			ptrTmpImg->m_pBitmap = new Bitmap(srcImage.dimx(), srcImage.dimy(), PixelFormat32bppPARGB);
			RGBQUAD *pBits = (RGBQUAD*)(ptrTmpImg->GetBits());
			int w = srcImage.dimx();
			int h = srcImage.dimy();
			int x,y,v0,v1,v2;
			
			if (srcImage.dimv() > 0) v0 = 0;
			if (srcImage.dimv() > 1) v1 = 1; else v1 = v0;
			if (srcImage.dimv() > 2) v2 = 2; else v2 = v1;

			for ( y = 0; y < h; ++y )
			{
				for ( x = 0; x < w; ++x )
				{
					pBits->rgbRed = int(srcImage.at(x,y,0,v0) + 0.5);
					pBits->rgbGreen = int(srcImage.at(x,y,0,v1) + 0.5);
					pBits->rgbBlue = int(srcImage.at(x,y,0,v2) + 0.5);
					pBits->rgbReserved = 255;
					++pBits;
				}
			}
			ptrTmpImg->FreeBits();
			ptrTmpImg->Save(strImageFileName.c_str());
			delete ptrTmpImg;
		}

	public:
		static int ReadStringSet(string strFileName, vector<string> &StringSet)
		{
			ifstream	inStringSet(strFileName.c_str( ));
			string		line;
			int			Count;

			inStringSet>>Count;
			getline(inStringSet, line, '\n');
			StringSet.resize(Count);
			for ( int t = 0; t < Count; t++ )
			{
				getline(inStringSet, StringSet[t], '\n');
			}

			inStringSet.clear( );
			inStringSet.close( );
			return	1;
		}

		static int WriteStringSet(string strFileName, vector<string> &StringSet)
		{
			ofstream	outStringSet(strFileName.c_str( ));
			int			Count	= (int) StringSet.size( );

			outStringSet<<Count<<endl;
			for ( int t = 0; t < Count; t++ )
			{
				outStringSet<<StringSet[t].c_str( )<<endl;
			}

			outStringSet.close( );
			outStringSet.clear( );
			return	1;
		}

		template <typename T>
		static int ReadValueSet(string strFileName, vector<T> &ValueSet)
		{
			ifstream	inValueSet(strFileName.c_str( ));
			int			Count;

			inValueSet>>Count;
			ValueSet.resize(Count);
			for ( int t = 0; t < Count; t++ )
			{
				inValueSet>>ValueSet[t];
			}

			inValueSet.clear( );
			inValueSet.close( );
			return	1;
		}

		template <typename T>
		static int WriteValueSet(string strFileName, vector<T> &ValueSet)
		{
			ofstream	outValueSet(strFileName.c_str( ));
			int			Count	= (int) ValueSet.size( );

			outValueSet.precision(6);
			outValueSet<<Count<<endl;
			for ( int t = 0; t < Count; t++ )
			{
				outValueSet<<setw(12)<<ValueSet[t]<<endl;
			}

			outValueSet.clear( );
			outValueSet.close( );
			return	1;
		}

		static	int NormalizeVector(vector<double> &ValueSet, double Norm = 1)
		{
			double	SumScore	= 0;
			int		Count		= (int) ValueSet.size( );

			for ( int t = 0; t < Count; t++ )
			{
				SumScore	+= ValueSet[t];
			}
			SumScore	/= Norm;
			for ( int t = 0; t < Count; t++ )
			{
				ValueSet[t]	/= SumScore;
			}
			return	1;
		}

		static	int ReadColors(string strColorFileName, vector<Color> &Colors)
		{
			ifstream	inColor(strColorFileName.c_str( ));
			int			r, g, b;
			int			Count;

			inColor>>Count;
			Colors.clear( );
			for ( int i = 0; i < Count; i++ )
			{
				inColor>>r>>g>>b;
				Colors.push_back(Color(r, g, b));
			}

			inColor.close( );
			inColor.clear( );
			return	1;
		}

		static double OriDiff(double ori1, double ori2)				// [-PI, PI]
		{
			double	ori	= ori2 - ori1;

			while ( ori > M_PI )
			{
				ori -= M_PI * 2;
			}

			while ( ori < - M_PI )
			{
				ori	+= M_PI * 2;
			}

			return	ori;
		}

		static int DistTrans1DX(TMatrix<double> &SrcMat, TMatrix<double> &TarMat, TMatrix<int> &SrcX, 
								int Y, int W, int PosX, double ax, double bx, int mov)
		{
			int	s1	= 0;
			int	s2	= W - 1;
			for ( int d = 0; d < W; d++ )
			{
				int		md		= d + PosX;
				int		mins	= __max(s1, md - mov + 1);
				int		maxs	= __min(s2, md + mov - 1);
				int		s		= __min(mins, maxs);
				double	maxm	= SrcMat[Y][s] - ax * (md-s) * (md-s) - bx * (md-s);

				for ( int p = mins + 1; p <= maxs; p++ )
				{
					double	m	= SrcMat[Y][p] - ax * (md-p) * (md-p) - bx * (md-p);
					if (m > maxm)
					{
						maxm	= m;
						s		= p;
					}
				}

				TarMat[Y][d]	= maxm;
				SrcX[Y][d]		= s;
			}

			return	1;
		}

		static int DistTrans1DY(TMatrix<double> &SrcMat, TMatrix<double> &TarMat, TMatrix<int> &SrcY, 
							    int X, int H, int PosY, double ay, double by, int mov)
		{
			int	s1	= 0;
			int	s2	= H - 1;
			for ( int d = 0; d < H; d++ )
			{
				int		md		= d + PosY;
				int		mins	= __max(s1, md - mov + 1);
				int		maxs	= __min(s2, md + mov - 1);
				int		s		= __min(mins, maxs);
				double	maxm	= SrcMat[s][X] - ay * (md-s) * (md-s) - by * (md-s);

				for ( int p = mins + 1; p <= maxs; p++ )
				{
					double	m	= SrcMat[p][X] - ay * (md-p) * (md-p) - by * (md-p);
					if (m > maxm)
					{
						maxm	= m;
						s		= p;
					}
				}

				TarMat[d][X]	= maxm;
				SrcY[d][X]		= s;
			}

			return	1;
		}

		static int DistanceTransform(TMatrix<double> &SrcMat, TMatrix<double> &TarMat, TMatrix<int> &SrcX, TMatrix<int> &SrcY,
									 int W, int H, int PosX, int PosY, 
									 double ax, double bx, double ay, double by, 
									 int MaxMovX, int MaxMovY)
		{
			TMatrix<double>		TransMat;
			TMatrix<int>		TransSrcX, TransSrcY;
			int					ret;

			TransSrcX.alloc(H, W);
			TransSrcY.alloc(H, W);
			TransMat.alloc(H, W);

			for ( int y = 0; y < H; y++ )
			{
				ret	= DistTrans1DX(SrcMat, TransMat, TransSrcX, y, W, PosX, ax, bx, MaxMovX);
			}

			for ( int x = 0; x < W; x++ )
			{
				ret	= DistTrans1DY(TransMat, TarMat, TransSrcY, x, H, PosY, ay, by, MaxMovY);
			}

			for ( int y = 0; y < H; y++ )
			{
				for ( int x = 0; x < W; x++ )
				{
					SrcY[y][x]	= TransSrcY[y][x];
					SrcX[y][x]	= TransSrcX[TransSrcY[y][x]][x];
				}
			}

			return	1;
		}

		static float GetRectOverlap(CRect &Rect1, CRect &Rect2, int OverlapType)
		{
			float	Overlap;

			if (OverlapType == 0)
			{
				int	S1	= Rect1.Height( ) * Rect1.Width( );
				int	S2	= Rect2.Height( ) * Rect2.Width( );
				int	H	= __max(0, __min(Rect1.bottom, Rect2.bottom) - __max(Rect1.top, Rect2.top));
				int	W	= __max(0, __min(Rect1.right, Rect2.right) - __max(Rect1.left, Rect2.left));
				Overlap	= 1.0f * (H * W) / (S1 + S2 - H * W);
			}

			if (OverlapType == 1)
			{
				int	S	= Rect1.Height( ) * Rect1.Width( );
				int	H	= __max(0, __min(Rect1.bottom, Rect2.bottom) - __max(Rect1.top, Rect2.top));
				int	W	= __max(0, __min(Rect1.right, Rect2.right) - __max(Rect1.left, Rect2.left));
				Overlap	= 1.0f * (H * W) / S;
			}

			if (OverlapType == 2)
			{
				int	S	= Rect2.Height( ) * Rect2.Width( );
				int	H	= __max(0, __min(Rect1.bottom, Rect2.bottom) - __max(Rect1.top, Rect2.top));
				int	W	= __max(0, __min(Rect1.right, Rect2.right) - __max(Rect1.left, Rect2.left));
				Overlap	= 1.0f * (H * W) / S;
			}

			return	Overlap;
		}

		//static inline double GetConvScore(double *f1, double *f2, int len)
		//{
		//	double	score	= 0;
		//	for ( int t = 0; t < len; t++ )
		//	{
		//		score	+= (*f1++) * (*f2++);
		//	}
		//	return	score;
		//}

		static inline int GetIndexVector(vector<double> &Src, vector<double> &Tar, vector<int> &Index)
		{
			Tar.resize(Src.size( ));
			for ( int t = 0; t < (int) Tar.size( ); t++ )
			{
				Tar[t]	= Src[Index[t]];
			}
			return	1;
		}

		static inline int VisualizeHistogram(Graphics &graphics, vector<double> &Hist, int SizeX, int SizeY, int PosX, int PosY, vector<Color> &Colors)
		{
			SolidBrush		myBrush(Color(255, 255, 255));
			Pen				myPen(&myBrush, 2);
			int				BinCnt	= (int) Hist.size( );
			int				BinSize	= SizeX / BinCnt;
			int				PSize	= int(BinSize * 0.8);

			myPen.SetColor(Color(0, 0, 0));
			graphics.FillRectangle(&myBrush, PosX, PosY, SizeX, SizeY);
			graphics.DrawRectangle(&myPen, PosX, PosY, SizeX, SizeY);
			myBrush.SetColor(Color(255, 0, 0));
			for ( int t = 0; t < BinCnt; t++ )
			{
				int X	= PosX + t * BinSize + (BinSize - PSize) / 2;
				int	Y	= int(PosY + SizeY * (1 - Hist[t]));
				int	W	= PSize;
				int	H	= PosY + SizeY - Y;
				myBrush.SetColor(Colors[t % Colors.size( )]);
				graphics.FillRectangle(&myBrush, X, Y, W, H);
			}

			return	1;
		}
};
