#include "StdAfx.h"
#include "HoGExtractor.h"

CHoGExtractor::CHoGExtractor(void)
{
	FeaStep		= 0.1;
	Boundary	= 1.0;
	MarX		= g_Parameters.MarX;
	MarY		= g_Parameters.MarY;
	Dim			= 32;
	eps			= 1e-6;

	DirVec.clear( );
	DirVec.push_back(make_pair( 1.0000, 0.0000));
	DirVec.push_back(make_pair( 0.9397, 0.3420));
	DirVec.push_back(make_pair( 0.7660, 0.6428));
	DirVec.push_back(make_pair( 0.5000, 0.8660));
	DirVec.push_back(make_pair( 0.1736, 0.9848));
	DirVec.push_back(make_pair(-0.1736, 0.9848));
	DirVec.push_back(make_pair(-0.5000, 0.8660));
	DirVec.push_back(make_pair(-0.7660, 0.6428));
	DirVec.push_back(make_pair(-0.9397, 0.3420));
}

CHoGExtractor::~CHoGExtractor(void)
{
}

int	CHoGExtractor::GetHoGMap(CImg<double> &Im, int BinSize, CImg<double> &FeaMap, int &W_HoG, int &H_HoG)
{
	CImg<double>		GradHist;
	TMatrix<double>		CellNorm;
	int					ImW		= Im.dimx( );
	int					ImH		= Im.dimy( );
	int					W_Cell	= int(1.0 * ImW / BinSize + 0.5);
	int					H_Cell	= int(1.0 * ImH / BinSize + 0.5);
	int					W_Map	= __max(0, W_Cell - 2);
	int					H_Map	= __max(0, H_Cell - 2);
	int					W_Crop	= W_Cell * BinSize;
	int					H_Crop	= H_Cell * BinSize;

	W_HoG	= int(ceil((float) (W_Map + 2 * MarX) / BlkDimX)) * BlkDimX + BlkDimX;
	H_HoG	= int(ceil((float) (H_Map + 2 * MarY) / BlkDimY)) * BlkDimY + BlkDimY;

	CellNorm.alloc(H_Cell, W_Cell);
	GradHist.assign(W_Cell, H_Cell, 1, 18, 0);
	FeaMap.assign(W_HoG, H_HoG, 1, Dim, 0);
	for ( int cy = 1; cy < H_Crop - 1; cy++ )
	{
		for ( int cx = 1; cx < W_Crop - 1; cx++ )
		{
			// gradient of each color channel
			int		x	= __min(cx, ImW - 2);
			int		y	= __min(cy, ImH - 2);
			double	dx1	= Im.at(x+1, y, 0, 0) - Im.at(x-1, y, 0, 0);
			double	dy1	= Im.at(x, y+1, 0, 0) - Im.at(x, y-1, 0, 0);
			double	v1	= dx1 * dx1 + dy1 * dy1;
			double	dx2	= Im.at(x+1, y, 0, 1) - Im.at(x-1, y, 0, 1);
			double	dy2	= Im.at(x, y+1, 0, 1) - Im.at(x, y-1, 0, 1);
			double	v2	= dx2 * dx2 + dy2 * dy2;
			double	dx3	= Im.at(x+1, y, 0, 2) - Im.at(x-1, y, 0, 2);
			double	dy3	= Im.at(x, y+1, 0, 2) - Im.at(x, y-1, 0, 2);
			double	v3	= dx3 * dx3 + dy3 * dy3;

			// pick channel with strongest gradient
			double	dx	= dx1;
			double	dy	= dy1;
			double	v	= v1;
			if (v < v2)
			{
				v	= v2;
				dx	= dx2;
				dy	= dy2;
			}
			if (v < v3)
			{
				v	= v3;
				dx	= dx3;
				dy	= dy3;
			}

			// snap to one of 18 orientations
			double	BestV	= 0;
			int		BestOri	= 0;
			for ( int t = 0; t < 9; t++ )
			{
				double	dot	= DirVec[t].first * dx + DirVec[t].second * dy;
				if (dot > BestV)
				{
					BestV	= dot;
					BestOri	= t;
				} else if (-dot > BestV)
				{
					BestV	= - dot;
					BestOri	= t + 9;
				}
			}

			double	cellX	= 1.0 * (cx + 0.5) / BinSize - 0.5;
			double	cellY	= 1.0 * (cy + 0.5) / BinSize - 0.5;
			int		PosX	= int(floor(cellX));
			int		PosY	= int(floor(cellY));
			double	px1		= cellX - PosX;
			double	py1		= cellY - PosY;
			double	px2		= 1.0 - px1;
			double	py2		= 1.0 - py1;

			BestV	= sqrt(v);
			if ((PosX >= 0) && (PosY >= 0))
			{
				GradHist.at(PosX, PosY, 0, BestOri)		+= BestV * px2 * py2;
			}
			if ((PosX >= 0) && (PosY+1 < H_Cell))
			{
				GradHist.at(PosX, PosY+1, 0, BestOri)	+= BestV * px2 * py1;
			}
			if ((PosX+1 < W_Cell) && (PosY >= 0))
			{
				GradHist.at(PosX+1, PosY, 0, BestOri)	+= BestV * px1 * py2;
			}
			if ((PosX+1 < W_Cell) && (PosY+1 < H_Cell))
			{
				GradHist.at(PosX+1, PosY+1, 0, BestOri)	+= BestV * px1 * py1;
			}
		}
	}

	for ( int t = 0; t < 9; t++ )
	{
		for ( int y = 0; y < H_Cell; y++ )
		{
			for ( int x = 0; x < W_Cell; x++ )
			{
				double	Score	= GradHist.at(x,y,0,t) + GradHist.at(x,y,0,t+9);
				CellNorm[y][x]	+= Score * Score;
			}
		}
	}

	for ( int y = 0; y < H_Map; y++ )
	{
		for ( int x = 0; x < W_Map; x++ )
		{
			double	n1	= 1.0 / sqrt(CellNorm[y+1][x+1] + CellNorm[y+1][x+2] + CellNorm[y+2][x+1] + CellNorm[y+2][x+2] + eps);
			double	n2	= 1.0 / sqrt(CellNorm[y+0][x+1] + CellNorm[y+0][x+2] + CellNorm[y+1][x+1] + CellNorm[y+1][x+2] + eps);
			double	n3	= 1.0 / sqrt(CellNorm[y+1][x+0] + CellNorm[y+1][x+1] + CellNorm[y+2][x+0] + CellNorm[y+2][x+1] + eps);
			double	n4	= 1.0 / sqrt(CellNorm[y+0][x+0] + CellNorm[y+0][x+1] + CellNorm[y+1][x+0] + CellNorm[y+1][x+1] + eps);
			double	t1	= 0;
			double	t2	= 0;
			double	t3	= 0;
			double	t4	= 0;

			// contrast-sensitive features
			for ( int t = 0; t < 18; t++)
			{
				double	s	= GradHist.at(x+1,y+1,0,t);
				double	h1	= __min(0.2, s * n1);
				double	h2	= __min(0.2, s * n2);
				double	h3	= __min(0.2, s * n3);
				double	h4	= __min(0.2, s * n4);

				t1	+= h1;
				t2	+= h2;
				t3	+= h3;
				t4	+= h4;
				FeaMap.at(x + MarX + 1, y + MarY + 1, 0, t)	= 0.5 * (h1 + h2 + h3 + h4);
				FeaMap.at(x + MarX + 1, y + MarY + 1, 0, t)	-= FeaStep;
			}

			// contrast-insensitive features
			for ( int t = 0; t < 9; t++)
			{
				double	s	= GradHist.at(x+1,y+1,0,t) + GradHist.at(x+1,y+1,0,t+9);
				double	h1	= __min(0.2, s * n1);
				double	h2	= __min(0.2, s * n2);
				double	h3	= __min(0.2, s * n3);
				double	h4	= __min(0.2, s * n4);
				FeaMap.at(x + MarX + 1, y + MarY + 1, 0, t + 18)	= 0.5 * (h1 + h2 + h3 + h4);
				FeaMap.at(x + MarX + 1, y + MarY + 1, 0, t + 18)	-= FeaStep;
			}

			FeaMap.at(x + MarX + 1, y + MarY + 1, 0, 27)	= 0.2357 * t1 - FeaStep;
			FeaMap.at(x + MarX + 1, y + MarY + 1, 0, 28)	= 0.2357 * t2 - FeaStep;
			FeaMap.at(x + MarX + 1, y + MarY + 1, 0, 29)	= 0.2357 * t3 - FeaStep;
			FeaMap.at(x + MarX + 1, y + MarY + 1, 0, 30)	= 0.2357 * t4 - FeaStep;
			FeaMap.at(x + MarX + 1, y + MarY + 1, 0, 31)	= 0.0;
		}
	}

	// boundary features
	for ( int y = 0; y < H_HoG; y++ )
	{
		for ( int x = 0; x < W_HoG; x++ )
		{
			if ((x <= MarX) || (y <= MarY) || (x > W_Map + MarX) || (y > H_Map + MarY))
			{
				FeaMap.at(x, y, 0, 31)	= Boundary;
			}
		}
	}

	return	1;
}
